from typing import Optional
from datetime import datetime
from pydantic import BaseModel


class FeatureBase(BaseModel):
    key: str
    value: str
    arm_id: int
    description: Optional[str] = None


class FeatureCreate(FeatureBase):
    pass


class FeatureUpdate(FeatureBase):
    pass


class FeatureInDBBase(FeatureBase):
    id: int
    time_created: datetime
    time_updated: datetime

    class Config:
        orm_mode = True


class Feature(FeatureInDBBase):
    pass


class FeatureInDB(FeatureInDBBase):
    pass



from typing import List, Optional

from fastapi.encoders import jsonable_encoder
from sqlalchemy.orm import Session

from abing.crud.base import CRUDBase

from abing.db.base import Arm, Allocation, Experiment
from abing.schemas.arm import ArmCreate, ArmUpdate


class CRUDArm(CRUDBase[Arm, ArmCreate, ArmUpdate]):
    def get_list(
        self, db: Session, *, skip: int = 0, limit: int = 100, experiment_id: int = None
    ) -> List[Arm]:
        query = db.query(self.model)
        if experiment_id:
            query = query.filter(self.model.experiment_id == experiment_id)
        return query.offset(skip).limit(limit).all()

    def get_from_experiment(
        self, db: Session, *, user_id: int, experiment: Experiment
    ) -> Optional[Arm]:
        return (
            db.query(self.model)
            .join(Allocation)
            .filter(Allocation.user_id == user_id)
            .join(Experiment)
            .filter(Experiment.id == experiment.id)
            .first()
        )


arm = CRUDArm(Arm)



from typing import TYPE_CHECKING

from sqlalchemy import Column, Integer, String, ForeignKey, DateTime
from sqlalchemy.orm import relationship
from sqlalchemy.sql import func

from abing.db.base_class import Base

if TYPE_CHECKING:
    from .experiment import Experiment  # noqa: F401
    from .feature import Feature  # noqa: F401
    from .allocation import Allocation


class Arm(Base):

    __tablename__ = "arms"

    id = Column(Integer, primary_key=True, index=True, autoincrement=True)
    name = Column(String(150), index=True)
    # NOTE: 실험 런칭 시 반영되는 가중치(e.g A가 100이고 B가 20이면 A는 100/120으로 트래픽 전달, B는 20/120으로 트래픽 전달)
    traffic_weight = Column(Integer, nullable=False)
    experiment_id = Column(Integer, ForeignKey("experiments.id"), nullable=False)
    experiment = relationship("Experiment", back_populates="arms")
    features = relationship("Feature", back_populates="arm", cascade="all,delete")
    allocations = relationship("Allocation", back_populates="arm", cascade="all,delete")
    time_created = Column(DateTime(timezone=True), server_default=func.now())
    time_updated = Column(
        DateTime(timezone=True), server_default=func.now(), onupdate=func.now()
    )



"""Change experiment schedule data type

Revision ID: 9f766facc4ab
Revises: 09746a50372f
Create Date: 2021-01-08 11:03:43.274138

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "9f766facc4ab"
down_revision = "09746a50372f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("experiments", "start_date", type_=sa.Date)
    op.alter_column("experiments", "end_date", type_=sa.Date)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("experiments", "start_date", type_=sa.DateTime)
    op.alter_column("experiments", "end_date", type_=sa.DateTime)
    # ### end Alembic commands ###



from typing import Any, Optional, List
from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.orm import Session

from abing import crud, models, schemas
from abing.routes import deps

from abing.services.route_service import HashRouter
from abing.core.exceptions import InvalidWeightException

router = APIRouter()


@router.get("/", response_model=List[schemas.Experiment])
async def get_experiments(
    db: Session = Depends(deps.get_db),
    skip: int = 0,
    limit: int = 100,
    current_user: models.User = Depends(deps.get_current_active_user),
):
    return crud.experiment.get_list(db, skip=skip, limit=limit)


@router.get("/route", response_model=List[schemas.Arm])
async def select_arms_by_experiments(
    db: Session = Depends(deps.get_db), *, user_id: str
):
    traffic_router = HashRouter(user_id)
    experiments = crud.experiment.get_list_by_running_state(db, is_running=True)
    selected_arms = list()

    if not experiments:
        raise HTTPException(
            status_code=404, detail="There are no experiments in progress."
        )

    for experiment in experiments:
        if len(experiment.arms) < 2:
            continue

        arm = crud.arm.get_from_experiment(
            db=db, user_id=user_id, experiment=experiment
        )
        if not arm:
            try:
                arm = traffic_router.route(experiment)
                crud.allocation.create(
                    db=db, obj_in={"user_id": user_id, "arm_id": arm.id}
                )
            except InvalidWeightException:
                continue

        selected_arms.append(arm)

    return selected_arms


@router.get("/{id}", response_model=schemas.Experiment)
async def get_experiment(
    id: int,
    db: Session = Depends(deps.get_db),
    current_user: models.User = Depends(deps.get_current_active_user),
) -> Any:
    return crud.experiment.get(db=db, id=id)


@router.get("/{id}/route", response_model=schemas.Arm)
async def select_arm_by_experiment(
    id: int, user_id: str, db: Session = Depends(deps.get_db)
):
    experiment = crud.experiment.get(db=db, id=id)

    if not experiment:
        raise HTTPException(status_code=404, detail="Experiment not found")
    elif len(experiment.arms) < 2:
        raise HTTPException(
            status_code=404,
            detail="Experiment arms are not set. Must required arm list at least 2.",
        )

    arm = crud.arm.get_from_experiment(db=db, user_id=user_id, experiment=experiment)
    if not arm:
        traffic_router = HashRouter(user_id)
        arm = traffic_router.route(experiment)
        crud.allocation.create(db=db, obj_in={"user_id": user_id, "arm_id": arm.id})
    return arm


@router.post("/")
async def create_experiment(
    experiment_in: schemas.ExperimentCreate,
    db: Session = Depends(deps.get_db),
    current_user: models.User = Depends(deps.get_current_active_user),
):
    return crud.experiment.create(db=db, obj_in=experiment_in)


@router.put("/{id}")
def update_experiment(
    *,
    db: Session = Depends(deps.get_db),
    id: int,
    experiment_in: schemas.ExperimentUpdate,
    current_user: models.User = Depends(deps.get_current_active_user),
) -> Any:

    experiment = crud.experiment.get(db=db, id=id)
    if not experiment:
        raise HTTPException(status_code=404, detail="Experiment not found")
    experiment = crud.experiment.update(db=db, db_obj=experiment, obj_in=experiment_in)
    return experiment


@router.delete("/{id}")
def delete_experiment(
    *,
    db: Session = Depends(deps.get_db),
    id: int,
    current_user: models.User = Depends(deps.get_current_active_user),
) -> Any:

    experiment = crud.experiment.get(db=db, id=id)
    if not experiment:
        raise HTTPException(status_code=404, detail="Experiment not found")
    experiment = crud.experiment.remove(db=db, id=id)
    return {"message": f"Experiment {id} is removed!", "type": "arm", "id": id}



from typing import TYPE_CHECKING

from sqlalchemy import Column, Integer, String, Boolean, DateTime, Date
from sqlalchemy.orm import relationship
from sqlalchemy.sql import func

from abing.db.base_class import Base

if TYPE_CHECKING:
    from .arm import Arm  # noqa: F401


class Experiment(Base):

    __tablename__ = "experiments"

    id = Column(Integer, primary_key=True, index=True, autoincrement=True)
    name = Column(String(200), index=True)
    is_running = Column(Boolean(), default=False)
    description = Column(String)
    start_date = Column(Date)
    end_date = Column(Date)
    # NOTE: Arm 순서는 Immutable해야 함.
    arms = relationship(
        "Arm", order_by="Arm.id", back_populates="experiment", cascade="all,delete"
    )
    time_created = Column(DateTime(timezone=True), server_default=func.now())
    time_updated = Column(
        DateTime(timezone=True), server_default=func.now(), onupdate=func.now()
    )



import os

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from abing.core.config import settings


engine = create_engine(settings.SQLALCHEMY_DATABASE_URI, pool_pre_ping=True)
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)



from typing import Any, Optional, List
from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.orm import Session

from abing import crud, models, schemas
from abing.routes import deps

router = APIRouter()


@router.get("/")
async def get_features(
    db: Session = Depends(deps.get_db),
    skip: int = 0,
    limit: int = 100,
    arm_id: Optional[int] = None,
    current_user: models.User = Depends(deps.get_current_active_user),
):
    return crud.feature.get_list(db, skip=skip, limit=limit, arm_id=arm_id)


@router.get("/{id}")
async def get_feature(
    id: int,
    db: Session = Depends(deps.get_db),
    current_user: models.User = Depends(deps.get_current_active_user),
) -> Any:
    return crud.feature.get(db=db, id=id)


@router.post("/")
async def create_feature(
    feature_in: schemas.FeatureCreate,
    db: Session = Depends(deps.get_db),
    current_user: models.User = Depends(deps.get_current_active_user),
):
    arm = crud.arm.get(db=db, id=feature_in.arm_id)
    if not arm:
        raise HTTPException(status_code=404, detail="Arm not found")
    return crud.feature.create(db=db, obj_in=feature_in)


@router.post("/bulk", response_model=List[schemas.Feature])
async def create_features(
    features_in: List[schemas.FeatureCreate],
    db: Session = Depends(deps.get_db),
    current_user: models.User = Depends(deps.get_current_active_user),
):

    created_features = list()
    for feature_in in features_in:
        arm = crud.arm.get(db=db, id=feature_in.arm_id)

        if not arm:
            raise HTTPException(status_code=404, detail="arm not found")

        feature = crud.feature.create(db=db, obj_in=feature_in)
        created_features.append(feature)
    return created_features


@router.put("/{id}")
def update_feature(
    *,
    db: Session = Depends(deps.get_db),
    id: int,
    feature_in: schemas.FeatureUpdate,
    current_user: models.User = Depends(deps.get_current_active_user),
) -> Any:

    feature = crud.feature.get(db=db, id=id)
    if not feature:
        raise HTTPException(status_code=404, detail="Feature not found")
    feature = crud.feature.update(db=db, db_obj=feature, obj_in=feature_in)
    return feature


@router.delete("/{id}")
def delete_feature(
    *,
    db: Session = Depends(deps.get_db),
    id: int,
    current_user: models.User = Depends(deps.get_current_active_user),
) -> Any:

    feature = crud.feature.get(db=db, id=id)
    arm_id = feature.arm_id
    if not feature:
        raise HTTPException(status_code=404, detail="Feature not found")
    feature = crud.feature.remove(db=db, id=id)
    return {"feature_id": id, "arm_id": arm_id, "message": f"Feature {id} is removed"}



import os
import click
import alembic.config as alembic_config
from abing.db.session import SessionLocal
from abing import crud, schemas


@click.group()
def abing():
    pass


@abing.command("initdb", short_help="initialze database.")
def initdb_command():
    config_file = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "alembic.ini"
    )
    alembic_config.main(argv=["--raiseerr", "-c", config_file, "upgrade", "head"])


@abing.command("createsuperuser", short_help="Create superuser")
@click.option("--email", prompt="email", help="User email")
@click.option("--password", prompt=True, confirmation_prompt=True, hide_input=True)
def superuser_command(email, password):
    db = SessionLocal()
    user_in = schemas.UserCreate(email=email, password=password, is_superuser=True)
    crud.user.create(db, obj_in=user_in)



from typing import TYPE_CHECKING

from sqlalchemy import Column, Integer, String, ForeignKey, DateTime
from sqlalchemy.orm import relationship
from sqlalchemy.sql import func

from abing.db.base_class import Base

if TYPE_CHECKING:
    from .arm import Arm  # noqa: F401


class Allocation(Base):

    __tablename__ = "allocations"

    id = Column(Integer, primary_key=True, index=True, autoincrement=True)
    arm_id = Column(Integer, ForeignKey("arms.id"), nullable=False)
    arm = relationship("Arm", back_populates="allocations")
    user_id = Column(String, nullable=False)
    time_created = Column(DateTime(timezone=True), server_default=func.now())
    time_updated = Column(
        DateTime(timezone=True), server_default=func.now(), onupdate=func.now()
    )



from typing import List

from fastapi.encoders import jsonable_encoder
from sqlalchemy.orm import Session

from abing.crud.base import CRUDBase

from abing.db.base import Experiment
from abing.schemas.experiment import ExperimentCreate, ExperimentUpdate


class CRUDExperiment(CRUDBase[Experiment, ExperimentCreate, ExperimentUpdate]):
    def get_list(
        self, db: Session, *, skip: int = 0, limit: int = 100
    ) -> List[Experiment]:
        return (
            db.query(self.model)
            .order_by(
                self.model.is_running.desc(),
                self.model.time_updated.desc(),
                self.model.time_created.desc(),
            )
            .offset(skip)
            .limit(limit)
            .all()
        )

    def get_list_by_running_state(self, db: Session, *, is_running: bool):
        return db.query(self.model).filter(self.model.is_running == is_running).all()


experiment = CRUDExperiment(Experiment)



"""Sampling Service
"""
from enum import Enum
from typing import List
import hashlib
import xxhash

from abing.models.experiment import Experiment
from abing.models.arm import Arm

from abing.core.exceptions import InvalidWeightException


class HashType(str, Enum):
    sha1 = "sha1"
    sha512 = "sha512"
    md5 = "md5"
    xxh32 = "xxh32"
    xxh64 = "xxh64"
    xxh128 = "xxh128"


class HashRouter:
    """Usage:
    >>> router = HashRouter('<User ID>')
    >>> exp = Session.query(Experiment).first()
    >>> selected_arm = router.route(exp)
    """

    def __init__(self, user_id: str, hash_type: HashType = HashType.xxh128):
        self.user_id = user_id
        if "xxh" in hash_type:
            self.hash = getattr(xxhash, hash_type)
        else:
            self.hash = getattr(hashlib, hash_type)

    def _digest_hash(self, data: str) -> str:
        """Hash digest"""
        h = self.hash()
        h.update(data.encode())
        hashed_data = h.hexdigest()
        return hashed_data

    def _get_hashed_ratio(self, userid: str, salt: str) -> float:
        """Hash Based Sampling
        Reference: https://tools.ietf.org/html/rfc5475
        """
        data = f"{userid}{salt}"
        hashed_data = self._digest_hash(data)
        max_hash_size = 16 ** len(hashed_data)
        hash_size = int(hashed_data, 16)
        return hash_size / max_hash_size

    def _pull_arm(self, ratio: float, arms: List[Arm]) -> Arm:
        total_weight = 0
        # NOTE: traffic weight data caching
        weight_table = dict()

        for arm in arms:
            weight_table[arm] = arm.traffic_weight
            total_weight += weight_table[arm]

        if total_weight <= 0:
            raise InvalidWeightException("Experiment weight should bigger than 0")

        current_weight = 0
        for arm in arms:
            try:
                next_weight = current_weight + (weight_table[arm] / total_weight)
            except ZeroDivisionError:
                next_weight = current_weight

            if current_weight <= ratio < next_weight:
                return arm

            current_weight = next_weight

        return arm

    def route(self, experiment: Experiment) -> Arm:
        """route user arm"""
        ratio = self._get_hashed_ratio(self.user_id, experiment.id)
        selected_arm = self._pull_arm(ratio, experiment.arms)
        return selected_arm



from datetime import datetime
from pydantic import BaseModel


class AllocationBase(BaseModel):
    arm_id: int
    user_id: str


class AllocationCreate(AllocationBase):
    pass


class AllocationUpdate(AllocationBase):
    pass


class AllocationInDBBase(AllocationBase):
    id: int
    time_created: datetime
    time_updated: datetime

    class Config:
        orm_mode = True


class Allocation(AllocationInDBBase):
    pass


class AllocationInDB(AllocationInDBBase):
    pass



"""Add Columns created and updated time

Revision ID: 88b3cab99dfd
Revises: ce5a475e1888
Create Date: 2020-12-19 07:46:02.653996

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "88b3cab99dfd"
down_revision = "ce5a475e1888"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "arms",
        sa.Column(
            "time_created",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
    )
    op.add_column(
        "arms",
        sa.Column(
            "time_updated",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
    )
    op.add_column(
        "experiments",
        sa.Column(
            "time_created",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
    )
    op.add_column(
        "experiments",
        sa.Column(
            "time_updated",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
    )
    op.add_column(
        "features",
        sa.Column(
            "time_created",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
    )
    op.add_column(
        "features",
        sa.Column(
            "time_updated",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("features", "time_updated")
    op.drop_column("features", "time_created")
    op.drop_column("experiments", "time_updated")
    op.drop_column("experiments", "time_created")
    op.drop_column("arms", "time_updated")
    op.drop_column("arms", "time_created")
    # ### end Alembic commands ###



from typing import Optional, Dict, Any

import secrets
from pydantic import BaseSettings, PostgresDsn, validator


class Settings(BaseSettings):

    PROJECT_NAME: str = "abing"
    API_VERSION: str = "/api/v1"

    SECRET_KEY: str = secrets.token_urlsafe(32)
    ACCESS_TOKEN_EXPIRE_MINUTES: int = 60 * 24 * 8
    TOKEN_SCHEME: str = "bearer"
    USERS_OPEN_REGISTRATION: bool = False

    ABING_DB_HOST: str
    ABING_DB_USER: str
    ABING_DB_PASSWORD: str
    ABING_DB: str
    ABING_DB_PORT: str = "5432"
    SQLALCHEMY_DATABASE_URI: Optional[PostgresDsn] = None

    SENTRY_DSN: Optional[str]

    @validator("SQLALCHEMY_DATABASE_URI", pre=True)
    def assemble_db_connection(
        cls, v: Optional[str], values: Dict[str, Any]  # noqa
    ) -> Any:
        if isinstance(v, str):
            return v
        return PostgresDsn.build(
            scheme="postgresql",
            user=values.get("ABING_DB_USER"),
            password=values.get("ABING_DB_PASSWORD"),
            host=values.get("ABING_DB_HOST"),
            port=values.get("ABING_DB_PORT"),
            path=f"/{values.get('ABING_DB', '')}",
        )

    class Config:
        case_sensitive = False


settings = Settings()



from typing import List, Optional
from datetime import datetime
from pydantic import BaseModel


class ArmBase(BaseModel):
    name: str
    traffic_weight: int


class ArmFeatureCreate(BaseModel):
    key: str
    value: str
    description: Optional[str] = None


class ArmCreate(ArmBase):
    experiment_id: int
    features: Optional[List[ArmFeatureCreate]]


class ArmUpdate(ArmBase):
    name: Optional[str]
    traffic_weight: Optional[int]


class ArmInDBBase(ArmBase):
    id: int
    experiment_id: int
    time_created: datetime
    time_updated: datetime

    class Config:
        orm_mode = True


class ArmFeature(BaseModel):
    key: str
    value: str

    class Config:
        orm_mode = True


class Arm(ArmInDBBase):
    features: Optional[List[ArmFeature]]
    pass


class ArmInDB(ArmInDBBase):
    pass



from datetime import timedelta
from typing import Any, Optional

from fastapi import APIRouter, Body, Depends, HTTPException
from fastapi.security import OAuth2PasswordRequestForm
from sqlalchemy.orm import Session

from abing import crud, models, schemas
from abing.routes import deps
from abing.core import security
from abing.core.config import settings
from abing.core.security import get_password_hash

router = APIRouter()


@router.post("/login/", response_model=schemas.Token)
def login_access_token(
    db: Session = Depends(deps.get_db), form_data: OAuth2PasswordRequestForm = Depends()
) -> Any:
    """
    OAuth2 compatible token login, get an access token for future requests
    """
    user = crud.user.authenticate(
        db, email=form_data.username, password=form_data.password
    )
    if not user:
        raise HTTPException(status_code=400, detail="Incorrect email or password")
    elif not crud.user.is_active(user):
        raise HTTPException(status_code=400, detail="Inactive user")
    access_token_expires = timedelta(minutes=settings.ACCESS_TOKEN_EXPIRE_MINUTES)
    return {
        "access_token": security.create_access_token(
            user.id, expires_delta=access_token_expires
        ),
        "token_type": settings.TOKEN_SCHEME,
    }



from typing import Any, Optional, List
from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.orm import Session

from abing import crud, models, schemas
from abing.routes import deps

router = APIRouter()


@router.get("/")
async def get_allocations(
    db: Session = Depends(deps.get_db),
    skip: int = 0,
    limit: int = 100,
    arm_id: Optional[int] = None,
    current_user: models.User = Depends(deps.get_current_active_user),
):
    return crud.allocation.get_list(db, skip=skip, limit=limit, arm_id=arm_id)


@router.get("/{id}")
async def get_allocation(
    id: int,
    db: Session = Depends(deps.get_db),
    current_user: models.User = Depends(deps.get_current_active_user),
) -> Any:
    return crud.allocation.get(db=db, id=id)


@router.delete("/{id}")
def delete_allocation(
    *,
    db: Session = Depends(deps.get_db),
    id: int,
    current_user: models.User = Depends(deps.get_current_active_user),
) -> Any:

    allocation = crud.allocation.get(db=db, id=id)
    arm_id = allocation.arm_id
    if not allocation:
        raise HTTPException(status_code=404, detail="allocation not found")
    allocation = crud.allocation.remove(db=db, id=id)
    return {
        "allocation_id": id,
        "arm_id": arm_id,
        "message": f"allocation {id} is removed",
    }



from typing import List

from sqlalchemy.orm import Session

from abing.crud.base import CRUDBase

from abing.db.base import Allocation
from abing.schemas.allocation import AllocationCreate, AllocationUpdate


class CRUDAllocation(CRUDBase[Allocation, AllocationCreate, AllocationUpdate]):
    def get_list(
        self, db: Session, *, skip: int = 0, limit: int = 100, arm_id: int = None
    ) -> List[Allocation]:
        query = db.query(self.model)
        if arm_id:
            query = query.filter(self.model.arm_id == arm_id)
        return query.offset(skip).limit(limit).all()


allocation = CRUDAllocation(Allocation)



from fastapi import APIRouter
from abing.routes.v1.endpoints import experiment, arm, feature, user, login, allocation

api_router = APIRouter()


api_router = APIRouter()
api_router.include_router(login.router, tags=["Login"])
api_router.include_router(user.router, prefix="/users", tags=["users"])
api_router.include_router(
    experiment.router, prefix="/experiments", tags=["Experiments"]
)
api_router.include_router(arm.router, prefix="/arms", tags=["Arms"])
api_router.include_router(feature.router, prefix="/features", tags=["Features"])
api_router.include_router(
    allocation.router, prefix="/allocations", tags=["Allocations"]
)



from typing import Any, Dict, Generic, List, Optional, Type, TypeVar, Union

from fastapi.encoders import jsonable_encoder
from pydantic import BaseModel
from sqlalchemy.orm import Session

from abing.db.base_class import Base

ModelType = TypeVar("ModelType", bound=Base)
CreateSchemaType = TypeVar("CreateSchemaType", bound=BaseModel)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseModel)


class CRUDBase(Generic[ModelType, CreateSchemaType, UpdateSchemaType]):
    def __init__(self, model: Type[ModelType]):
        """
        source: https://github.com/tiangolo/full-stack-fastapi-postgresql/blob/master/%7B%7Bcookiecutter.project_slug%7D%7D/backend/app/app/crud/base.py
        CRUD object with default methods to Create, Read, Update, Delete (CRUD).

        **Parameters**

        * `model`: A SQLAlchemy model class
        * `schema`: A Pydantic model (schema) class
        """
        self.model = model

    def get(self, db: Session, id: Any) -> Optional[ModelType]:
        return db.query(self.model).filter(self.model.id == id).first()

    def get_list(
        self, db: Session, *, skip: int = 0, limit: int = 100
    ) -> List[ModelType]:
        return db.query(self.model).offset(skip).limit(limit).all()

    def create(self, db: Session, *, obj_in: CreateSchemaType) -> ModelType:
        obj_in_data = jsonable_encoder(obj_in)
        db_obj = self.model(**obj_in_data)  # type: ignore
        db.add(db_obj)
        db.commit()
        db.refresh(db_obj)
        return db_obj

    def update(
        self,
        db: Session,
        *,
        db_obj: ModelType,
        obj_in: Union[UpdateSchemaType, Dict[str, Any]]
    ) -> ModelType:
        obj_data = jsonable_encoder(db_obj)
        if isinstance(obj_in, dict):
            update_data = obj_in
        else:
            update_data = obj_in.dict(exclude_unset=True)
        for field in obj_data:
            if field in update_data:
                setattr(db_obj, field, update_data[field])
        db.add(db_obj)
        db.commit()
        db.refresh(db_obj)
        return db_obj

    def remove(self, db: Session, *, id: int) -> ModelType:
        obj = db.query(self.model).get(id)
        db.delete(obj)
        db.commit()
        return obj



from .user import User, UserCreate, UserUpdate
from .token import Token, TokenPayload
from .experiment import Experiment, ExperimentCreate, ExperimentUpdate
from .arm import Arm, ArmCreate, ArmUpdate
from .feature import Feature, FeatureCreate, FeatureUpdate
from .allocation import Allocation, AllocationCreate, AllocationUpdate
from .message import Message



# Import all the models, so that Base has them before being
# imported by Alembic
from abing.db.base_class import Base  # noqa
from abing.models.user import User  # noqa
from abing.models.experiment import Experiment  # noqa
from abing.models.arm import Arm  # noqa
from abing.models.feature import Feature  # noqa
from abing.models.allocation import Allocation  # noqa



"""First migration

Revision ID: 16bd2982f46e
Revises: 
Create Date: 2020-12-04 12:37:43.821990

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "16bd2982f46e"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "experiments",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("name", sa.String(length=200), nullable=True),
        sa.Column("is_running", sa.Boolean(), nullable=True),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("start_date", sa.DateTime(), nullable=True),
        sa.Column("end_date", sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_experiments_id"), "experiments", ["id"], unique=False)
    op.create_index(op.f("ix_experiments_name"), "experiments", ["name"], unique=False)
    op.create_table(
        "arms",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("name", sa.String(length=150), nullable=True),
        sa.Column("experiment_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["experiment_id"],
            ["experiments.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_arms_id"), "arms", ["id"], unique=False)
    op.create_index(op.f("ix_arms_name"), "arms", ["name"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_arms_name"), table_name="arms")
    op.drop_index(op.f("ix_arms_id"), table_name="arms")
    op.drop_table("arms")
    op.drop_index(op.f("ix_experiments_name"), table_name="experiments")
    op.drop_index(op.f("ix_experiments_id"), table_name="experiments")
    op.drop_table("experiments")
    # ### end Alembic commands ###



from typing import List, Optional
from datetime import datetime, date
from pydantic import BaseModel
from .arm import Arm


class ExperimentBase(BaseModel):
    name: str
    description: Optional[str] = None
    is_running: Optional[bool] = False
    start_date: Optional[date]
    end_date: Optional[date]


class ExperimentCreate(ExperimentBase):
    pass


class ExperimentUpdate(ExperimentBase):
    name: Optional[str]
    description: Optional[str]
    is_running: Optional[bool]
    start_date: Optional[date]
    end_date: Optional[date]


class ExperimentInDBBase(ExperimentBase):
    id: int
    time_created: datetime
    time_updated: datetime

    class Config:
        orm_mode = True


class Experiment(ExperimentInDBBase):
    arms: Optional[List[Arm]]
    pass


class ExperimentInDB(ExperimentInDBBase):
    pass



"""Add feature model

Revision ID: ce5a475e1888
Revises: 16bd2982f46e
Create Date: 2020-12-19 07:39:35.572972

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "ce5a475e1888"
down_revision = "16bd2982f46e"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "features",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("key", sa.String(length=200), nullable=True),
        sa.Column("value", sa.String(), nullable=True),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("arm_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["arm_id"],
            ["arms.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_features_id"), "features", ["id"], unique=False)
    op.create_index(op.f("ix_features_key"), "features", ["key"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_features_key"), table_name="features")
    op.drop_index(op.f("ix_features_id"), table_name="features")
    op.drop_table("features")
    # ### end Alembic commands ###



"""Add traffic weight Column to Arm model

Revision ID: 582710b9fab6
Revises: 88b3cab99dfd
Create Date: 2020-12-20 05:29:49.834686

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "582710b9fab6"
down_revision = "88b3cab99dfd"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("arms", sa.Column("traffic_weight", sa.Integer(), nullable=False))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("arms", "traffic_weight")
    # ### end Alembic commands ###



from typing import Any, List, Optional
from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.orm import Session

from abing import crud, models, schemas
from abing.routes import deps

router = APIRouter()


@router.get("/", response_model=List[schemas.Arm])
async def get_arms(
    db: Session = Depends(deps.get_db),
    skip: int = 0,
    limit: int = 100,
    experiment_id: Optional[int] = None,
    current_user: models.User = Depends(deps.get_current_active_user),
):
    arm_list = crud.arm.get_list(
        db, skip=skip, limit=limit, experiment_id=experiment_id
    )
    return arm_list


@router.get("/{id}", response_model=schemas.Arm)
async def get_arm(
    id: int,
    db: Session = Depends(deps.get_db),
    current_user: models.User = Depends(deps.get_current_active_user),
) -> Any:
    return crud.arm.get(db=db, id=id)


@router.post("/", response_model=schemas.Arm)
async def create_arm(
    arm_in: schemas.ArmCreate,
    db: Session = Depends(deps.get_db),
    current_user: models.User = Depends(deps.get_current_active_user),
):
    experiment = crud.experiment.get(db=db, id=arm_in.experiment_id)
    if not experiment:
        raise HTTPException(status_code=404, detail="experiment not found")

    arm_in = arm_in.dict()
    features = arm_in.pop("features", [])

    arm = crud.arm.create(db=db, obj_in=arm_in)

    for feature in features:
        feature_in = {**feature, "arm_id": arm.id}
        feature_out = crud.feature.create(db=db, obj_in=feature_in)
    return arm


@router.post("/bulk", response_model=List[schemas.Arm])
async def create_arms(
    arms_in: List[schemas.ArmCreate],
    db: Session = Depends(deps.get_db),
    current_user: models.User = Depends(deps.get_current_active_user),
):

    created_arms = list()
    for arm_in in arms_in:
        experiment = crud.experiment.get(db=db, id=arm_in.experiment_id)

        if not experiment:
            raise HTTPException(status_code=404, detail="experiment not found")

        arm_in = arm_in.dict()
        features = arm_in.pop("features", []) or []
        arm = crud.arm.create(db=db, obj_in=arm_in)

        for feature in features:
            feature_in = {**feature, "arm_id": arm.id}
            feature_out = crud.feature.create(db=db, obj_in=feature_in)

        created_arms.append(arm)
    return created_arms


@router.put("/{id}")
def update_arm(
    *,
    db: Session = Depends(deps.get_db),
    id: int,
    arm_in: schemas.ArmUpdate,
    current_user: models.User = Depends(deps.get_current_active_user),
) -> Any:

    arm = crud.arm.get(db=db, id=id)
    if not arm:
        raise HTTPException(status_code=404, detail="arm not found")
    arm = crud.arm.update(db=db, db_obj=arm, obj_in=arm_in)
    return arm


@router.delete("/{id}")
def delete_arm(
    *,
    db: Session = Depends(deps.get_db),
    id: int,
    current_user: models.User = Depends(deps.get_current_active_user),
) -> Any:

    arm = crud.arm.get(db=db, id=id)
    experiment_id = arm.experiment_id
    if not arm:
        raise HTTPException(status_code=404, detail="arm not found")
    crud.arm.remove(db=db, id=id)
    return {
        "message": f"Arm {id} is removed!",
        "type": "arm",
        "arm_id": id,
        "experiment_id": experiment_id,
    }



import os
from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

home = os.path.abspath(os.path.dirname(__name__))

setup(
    name="abing",
    version="0.1.2",
    packages=find_packages(home),
    include_package_data=True,
    author="Dohyung Park",
    author_email="dohyung.prk@gmail.com",
    description="Simply setup AB testing environment",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dohyungp/abing",
    platforms=["any"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "fastapi",
        "uvicorn",
        "sqlalchemy",
        "alembic",
        "pydantic[email]",
        "python-jose[cryptography]",
        "passlib[bcrypt]",
        "psycopg2-binary",
        "xxhash",
        "python-multipart",
        "aiofiles",
        "jinja2",
        "click",
    ],
    extras_require={"sentry": ["sentry-sdk"]},
    entry_points={
        "console_scripts": [
            "abing=abing.cli:abing",
        ],
    },
    python_requires=">=3.6",
)



import os
from typing import Generator, Any, Optional
from fastapi import FastAPI, Request
from fastapi import Depends, HTTPException
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates

from starlette.middleware.cors import CORSMiddleware

from sqlalchemy.orm import Session

from abing.routes.v1.api import api_router
from abing.core.config import settings
from abing.core.exceptions import SentryNotFoundException

try:
    import sentry_sdk
    from sentry_sdk.integrations.asgi import SentryAsgiMiddleware
except (ModuleNotFoundError, ImportError):
    if settings.SENTRY_DSN is not None:
        # NOTE: future should change custom exception.
        raise SentryNotFoundException(
            "For using sentry, please install sentry sdk(simply you could install via pip install abing[sentry])."
        )

app = FastAPI(title=settings.PROJECT_NAME)

app.include_router(api_router, prefix=settings.API_VERSION)

directory = os.path.dirname(__file__)
app.mount("/static", StaticFiles(directory=f"{directory}/static/static"), name="static")
templates = Jinja2Templates(directory=f"{directory}/static")


@app.route("/{paths:path}")
async def serve(request: Request, paths: Optional[str] = None):
    return templates.TemplateResponse("index.html", {"request": request})


# NOTE: Allow CORS
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# NOTE: Sentry configuration should place in last.
if settings.SENTRY_DSN:
    sentry_sdk.init(
        dsn=settings.SENTRY_DSN,
        traces_sample_rate=1.0,
    )
    app = SentryAsgiMiddleware(app)



"""Add model Allocation

Revision ID: 235757c5bc34
Revises: 9f766facc4ab
Create Date: 2021-01-11 10:23:37.587900

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "235757c5bc34"
down_revision = "9f766facc4ab"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "allocations",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("arm_id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.String(), nullable=False),
        sa.Column(
            "time_created",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "time_updated",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["arm_id"],
            ["arms.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_allocations_id"), "allocations", ["id"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_allocations_id"), table_name="allocations")
    op.drop_table("allocations")
    # ### end Alembic commands ###



"""Add User model

Revision ID: 09746a50372f
Revises: 582710b9fab6
Create Date: 2020-12-23 03:05:32.652235

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "09746a50372f"
down_revision = "582710b9fab6"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "user",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(), nullable=True),
        sa.Column("email", sa.String(), nullable=False),
        sa.Column("password", sa.String(), nullable=False),
        sa.Column("is_active", sa.Boolean(), nullable=True),
        sa.Column("is_superuser", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_user_email"), "user", ["email"], unique=True)
    op.create_index(op.f("ix_user_id"), "user", ["id"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_user_id"), table_name="user")
    op.drop_index(op.f("ix_user_email"), table_name="user")
    op.drop_table("user")
    # ### end Alembic commands ###



from sqlalchemy import Boolean, Column, Integer, String
from sqlalchemy.orm import relationship

from abing.db.base_class import Base


class User(Base):
    id = Column(Integer, primary_key=True, index=True)
    name = Column(String)
    email = Column(String, unique=True, index=True, nullable=False)
    password = Column(String, nullable=False)
    is_active = Column(Boolean(), default=True)
    is_superuser = Column(Boolean(), default=False)
